/*
 * xrlogin - X based interface to rlogin, allowing interactive entry
 *           of hostname
 *
 * Copyright (c) 1991 by William LeFebvre
 *
 * This software is free and may be freely redistributed, but the author
 * requires that this entire comment area remain intact.
 */

#define VERSION 1
#include "patchlevel.h"

#define TRYHOST "./tryhost"

#define IP_ADDR_LEN  4

#define FlushDisplay      (XFlush(XtDisplay(context->toplevel)))

#include <stdio.h>
#include <fcntl.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <netdb.h>

/* includes for widgets go here */
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Viewport.h>

#include <X11/Xaw/Cardinals.h>

#include "xrlogin.h"

char copyright[] = "xrlogin, Copyright (c) 1991 by William LeFebvre";

extern int sys_nerr;
extern char *sys_errlist[];
extern int errno;

Widget wg_connect;

AppData app_data;

/* the context gets passed to the callbacks so that they have handles for
   all the other widgets, including the toplevel, and the application context.
 */

typedef struct Context {
    XtAppContext app_context;
    Widget       toplevel;
    Widget       host;
    Widget       message;
    Widget       telnet;
    Widget       user;
} *Context;

/* * * * MESSAGE HANDLING ROUTINES * * * */

/* show_message(wg, msg) - display "msg" in the message area */

XtIntervalId message_intid = NULL;

show_message(context, msg)

Context context;
char *msg;

{
    static char message_area[1024];
    static Arg nargs[] =
    {
	{XtNlabel, (XtArgVal)message_area},
    };

    /* make sure there isn't a timer pending */
    if (message_intid != NULL)
    {
	XtRemoveTimeOut(message_intid);
	message_intid = NULL;
    }

    strcpy(message_area, msg);
    XtSetValues(context->message, nargs, XtNumber(nargs));
}

clear_message(context, intid)

Context context;
XtIntervalId *intid;

{
    message_intid = NULL;
    show_message(context, "");
}

expire_message(context)

Context context;

{
    /* make sure the message expires after a certain period of time */
    /* 8 seconds ought to be good */
    message_intid = XtAppAddTimeOut(context->app_context,
				    8*1000, clear_message, context);
}


/* * * * * * * */

char *tryhost(name, ipaddr, timeout, ping)

char *name;
u_long ipaddr;
int timeout;
char *ping;

{
    union wait st;
    int pfd[2];
    register char *p;
    register int i;
    register int child;
    static char msg[1025];
    static char addr[20];

    /* convert timeout delay into a string (so we can pass it to execl) */
    sprintf(msg, "%d", timeout);

    /* do the same with the internet address */
    sprintf(addr, "%ld", ipaddr);

    /* determine type of ping to use and remember in "i" */
    if ((p = rindex(ping, '/')) == NULL)
    {
	p = ping;
    }
    else
    {
	p++;
    }
    i = strcmp(p, "tryhost") == 0;

    /* launch a "ping", capture the output, and look at the return status */
    pipe(pfd);
    if ((child = vfork()) == 0)
    {
	/* child process */
	dup2(pfd[1], 1);
	dup2(pfd[1], 2);
	close(pfd[0]);
	close(pfd[1]);
	execl(ping, p, i ? addr : name, msg, 0);
	_exit(100);
    }

    /* parent */
    close(pfd[1]);
    while ((i = wait(&st)) != child)
    {
	if (i == -1)
	{
	    close(pfd[0]);
	    return(sys_errlist[errno >= sys_nerr ? 0 : errno]);
	}
    }
    if (WIFEXITED(st) && st.w_retcode == 0)
    {
	close(pfd[0]);
	return(NULL);
    }
    else
    {
	i = read(pfd[0], msg, sizeof(msg)-1);
	close(pfd[0]);
	if (i == -1)
	{
	    return(sys_errlist[errno >= sys_nerr ? 0 : errno]);
	}
	msg[i] = '\0';
	if ((p = index(msg, '\n')) != NULL)
	{
	    *p = '\0';
	}
	return(msg);
    }
}

/* Callbacks for command widgets: */

cmd_quit()

{
    exit(0);
}

cmd_connect(w, context, call)

Widget w;
Context context;
caddr_t call;

{
    register int num;
    register int isrlogin;
    struct hostent *hent;
    char *reach;
    char *cmd;
    unsigned int addrparts[4];
    unsigned int address;
    char *opt1 = NULL;         /* optional argument to xterm */
    char *opt2 = NULL;         /* optional argument to xterm */
    static String str;
    static String name;
    static Arg nargs[] =
    {
	{XtNstring, NULL},
    };

    /* update the display */
    FlushDisplay;

    /* get the desired host name */
    nargs[0].value = (XtArgVal)&name;
    XtGetValues(context->host, nargs, XtNumber(nargs));

    /* if hostname is null, be friendly */
    if (name[0] == '\0')
    {
	show_message(context, "No host name specified!");
	return;
    }

    /* get the connection type, which will be the command to use */
    if ((cmd = XawToggleGetCurrent(context->telnet)) == NULL)
    {
	show_message(context, "No connection type selected.");
	return;
    }

    /* rlogin requires special handling */
    isrlogin = strcmp(cmd, "rlogin") == 0;

    /* is it just an IP address? */
    num = sscanf(name, "%u.%u.%u.%u",
		 &addrparts[0],
		 &addrparts[1],
		 &addrparts[2],
		 &addrparts[3]);
    if (num == 4)
    {
	/* internet number specified---convert to integer */
	address = (addrparts[0] << 24) |
	          (addrparts[1] << 16) |
		  (addrparts[2] << 8)  |
		  addrparts[3];

	/* use telnet since rlogin doesn't grok IP numbers */
	if (isrlogin)
	{
	    cmd = "telnet";
	    show_message(context, "Using telnet");
	    FlushDisplay;
	}
    }
    else
    {
	/* try to get its IP address */
	if ((hent = gethostbyname(name)) == NULL)
	{
	    show_message(context, "That host name is not known.");
	    return;
	}
	if (hent->h_length != IP_ADDR_LEN)
	{
	    show_message(context, "Unexpected address length");
	    return;
	}
	bcopy(hent->h_addr_list[0], &address, IP_ADDR_LEN);

	/* if rlogin, get and use username */
	if (isrlogin)
	{
	    nargs[0].value = (XtArgVal)&str;
	    XtGetValues(context->user, nargs, XtNumber(nargs));
	    if (str[0] != '\0')
	    {
		opt1 = "-l";
		opt2 = (char *)str;
	    }
	}
    }

    /* make sure it is up and reachable */
    show_message(context, "Connecting...");
    FlushDisplay;
    if ((reach = tryhost(name, address,
			 app_data.timeout,
			 app_data.ping)) != NULL)
    {
	show_message(context, reach);
    }
    else     /* everything is okay, initiate a connection */
    {
	if (vfork() == 0)
	{
	    /* child */
	    execl("/usr/bin/X11/xterm", "xterm",
		  "-T", name,
		  "-e", cmd, name, opt1, opt2, 0);
	    _exit(100);
	}

	/* set the message to expire */
	expire_message(context);
    }
}

cmd_select(w, context, list)

Widget w;
Context context;
XawListReturnStruct *list;

{
    static Arg nargs[] =
    {
	{XtNstring, NULL},
    };

    /* retrieve the selected hostname and jam it into host widget */
    nargs[0].value = (XtArgVal)(list->string);
    XtSetValues(context->host, nargs, XtNumber(nargs));

    /* unhighlight the widget */
    XawListUnhighlight(w);

    /* erase the message area (since the host widget changed) */
    show_message(context, "");
}

/* Other Callbacks: */

/* this callback is called whenever a text widget changes */

void widget_changed(widget, context, call)

Widget widget;
Context context;
caddr_t call;

{
    /* clear out message area */
    show_message(context, "");
}

/* Actions: */

act_connect(w, event, params, num_params)

Widget w;
XEvent *event;
String *params;
int    num_params;

{
    XtCallCallbacks(wg_connect, XtNcallback, NULL);
}

static XtActionsRec actionsTable[] =
{
    {"Connect", (XtActionProc)act_connect},
    {NULL, NULL}
};

/* Initialization stuff: */

/* init_textwidget:  set a text widget to call "widget_changed" when 
   it is changed */

init_textwidget(context, widget)

Context context;
Widget widget;

{
    static Widget source;
    static Arg warg[] =
    {
	{XtNtextSource, (XtArgVal)&source},
    };

    /* we have to retrieve the textsource widget and add the callback
       to that widget's callback list!  */

    XtGetValues(widget, warg, 1);
    XtAddCallback(source, XtNcallback, widget_changed, context);
}

int filesize(name)

char *name;

{
    struct stat st;

    return(stat(name, &st) == 0 ? (int)st.st_size : 0);
}

/* 
 *  build_hosts_list(filename) - read the specified file and build a list
 *        of hosts to choose from suitable for a List widget.
 */

char **build_hosts_list(filename)

char *filename;

{
    register int size;
    register int fd;
    register char *ptr;
    register char *tptr;
    register char **pptr;
    char *hostdata = NULL;
    char **hosts = NULL;
    static char buff[1024];

    /* if filename doesn't start with '/', prepend $HOME/ to it */
    if (filename[0] != '/')
    {
	strcpy(buff, getenv("HOME"));
	strcat(buff, "/");
	strcat(buff, filename);
	filename = buff;
    }

    if ((size = filesize(filename)) > 0)
    {
	if ((hostdata = (char *)malloc(size)) != NULL)
	{
	    if ((fd = open(filename, O_RDONLY)) >= 0)
	    {
		/* read the entire file */
		size = read(fd, hostdata, size);
		close(fd);

		/* determine number of lines */
		/* reuse "fd" as a temporary */
		fd = 0;
		ptr = hostdata;
		while (--size >= 0)
		{
		    if (*ptr++ == '\n')
		    {
			fd++;
		    }
		}

		/* guard against lines == 0 */
		if (fd > 0)
		{
		    /* allocate array of pointers */
		    if ((hosts = (char **)malloc((fd + 1) * sizeof(char *))) != NULL)
		    {
			/* build array of pointers */
			pptr = hosts;
			size = fd;
			ptr = hostdata;
			while (--size >= 0)
			{
			    /* put pointer to this line in array */
			    *pptr++ = ptr;

			    /* find end of line */
			    while (*ptr++ != '\n');

			    /* back up over trailing spaces */
			    tptr = ptr - 1;
			    while (*--tptr == ' ');

			    /* terminate string after last meaningful char */
			    *(tptr+1) = '\0';
			}
			*pptr = NULL;
		    }
		}
	    }
	}
    }

    /* if build was unsuccessful, free buffer */
    if (hosts == NULL && hostdata != NULL)
    {
	free(hostdata);
    }

    /* done */
    return(hosts);
}

main(argc, argv)

int argc;
char *argv[];

{
    XtAppContext app_con;
    XtWidgetGeometry wgeo;
    Widget toplevel;
    Widget wg_form, wg_label, wg_host, wg_quit, wg_message;
    Widget wg_type, wg_telnet, wg_rlogin, wg_hostlist, wg_hostview;
    Widget wg_user, wg_username;
    struct Context context;
    char **hosts;

    /* X initialization */
    toplevel = XtAppInitialize(&app_con, "XRlogin",
			       options, XtNumber(options),
			       &argc, argv, NULL,
			       NULL, ZERO);

    /* syntax check */
    if (argc > 1)
    {
	usage_message();
	exit(1);
    }

    /* add our specific actions */
    XtAppAddActions(app_con, actionsTable, XtNumber(actionsTable));

    /* get application resources */
    XtGetApplicationResources(toplevel,
			      &app_data,
			      resources,
			      XtNumber(resources),
			      NULL, ZERO);

    /* build the hosts list from the file in the home directory */
    hosts = build_hosts_list(app_data.hostfile);

    /* create all the widgets */
    wg_form = XtCreateManagedWidget("form", formWidgetClass, toplevel,
				    NULL, ZERO);

    wg_label = XtVaCreateManagedWidget("label", labelWidgetClass, wg_form,
				       NULL);

    wg_host = XtVaCreateManagedWidget("host", asciiTextWidgetClass, wg_form,
				      XtNfromHoriz, (XtArgVal)wg_label,
				      XtNeditType, (XtArgVal)XawtextEdit,
				      NULL);
    init_textwidget(&context, wg_host);

    wg_message = XtVaCreateManagedWidget("message", labelWidgetClass, wg_form,
					 XtNfromVert, (XtArgVal)wg_label,
					 NULL);

    wg_type = XtVaCreateManagedWidget("type", labelWidgetClass, wg_form,
				      XtNfromVert, (XtArgVal)wg_message,
				      NULL);

    wg_telnet = XtVaCreateManagedWidget("telnet", toggleWidgetClass, wg_form,
					XtNfromVert, (XtArgVal)wg_message,
					XtNfromHoriz, (XtArgVal)wg_type,
					NULL);

    wg_rlogin = XtVaCreateManagedWidget("rlogin", toggleWidgetClass, wg_form,
					XtNfromVert, (XtArgVal)wg_message,
					XtNfromHoriz, (XtArgVal)wg_telnet,
					XtNradioGroup, (XtArgVal)wg_telnet,
					XtNstate, (XtArgVal)True,
					NULL);

    wg_username = XtVaCreateManagedWidget("username", labelWidgetClass,
					   wg_form,
					   XtNfromVert, (XtArgVal)wg_type,
					   NULL);

    wg_user = XtVaCreateManagedWidget("user", asciiTextWidgetClass, wg_form,
				      XtNfromHoriz, (XtArgVal)wg_username,
				      XtNfromVert, (XtArgVal)wg_type,
				      XtNeditType, (XtArgVal)XawtextEdit,
				      NULL);

    wg_quit = XtVaCreateManagedWidget("quit", commandWidgetClass, wg_form,
				      XtNfromVert, (XtArgVal)wg_username,
				      NULL);

    wg_connect = XtVaCreateManagedWidget("connect", commandWidgetClass, wg_form,
					 XtNfromHoriz, (XtArgVal)wg_quit,
					 XtNfromVert, (XtArgVal)wg_username,
					 NULL);

    /* create list widget for hostname choices only if we need to */
    if (hosts != NULL)
    {
	/* create the viewport (with scrollbar) */
	wg_hostview = XtVaCreateManagedWidget("hostview",
					      viewportWidgetClass, wg_form,
#ifdef notdef
					      XtNfromHoriz, (XtArgVal)wg_message,
					      XtNfromVert, (XtArgVal)wg_host,
#endif
					      XtNallowVert, (XtArgVal)True,
					      NULL);

	/* create the list widget itself */
	wg_hostlist = XtVaCreateManagedWidget("hostlist",
					      listWidgetClass, wg_hostview,
					      XtNverticalList, (XtArgVal)True,
					      XtNlist, hosts,
					      XtNdefaultColumns, (XtArgVal)1,
					      XtNforceColumns, (XtArgVal)True,
					      NULL);

	/* add a callback for selection */
	XtAddCallback(wg_hostlist, XtNcallback, cmd_select, &context);

	/* get the intended width of list widget */
	XtQueryGeometry(wg_hostlist, NULL, &wgeo);

	/* add width sufficient for scrollbar */
	/* 14 is MAGIC:  should be taken from resources */
	/* set viewport width */
	XtVaSetValues(wg_hostview, XtNwidth, (XtArgVal)(wgeo.width + 14),
		      NULL);
    }

    /* build the context structure used by all the callbacks */
    context.app_context = app_con;
    context.toplevel = toplevel;
    context.host = wg_host;
    context.message = wg_message;
    context.telnet = wg_telnet;
    context.user = wg_user;

    XtAddCallback(wg_quit, XtNcallback, cmd_quit, NULL);
    XtAddCallback(wg_connect, XtNcallback, cmd_connect, &context);

    /* install the accelerators for the buttons */
    XtInstallAllAccelerators(wg_form, wg_connect);
    XtInstallAllAccelerators(wg_form, wg_quit);
    XtInstallAllAccelerators(wg_host, wg_connect);
    XtInstallAllAccelerators(wg_host, wg_quit);
    XtInstallAllAccelerators(wg_user, wg_connect);
    XtInstallAllAccelerators(wg_user, wg_quit);

    XtRealizeWidget(toplevel);
    XtAppMainLoop(app_con);
}

usage_message()

{
    fprintf(stderr, "\
xrlogin version %d.%d, Copyright (c) 1991 by William LeFebvre,\n\
xrlogin understands all X toolkit options plus:\n\
\t-hostfile file     file containing popular host names\n\
\t-timeout time      time to wait for remote end to answer\n",
	    VERSION, PATCHLEVEL);
}
